

import flask

from viggocore.common import exception, utils


def export_capabilities_and_policies(self, id):
    if flask.request.is_json:
        data = flask.request.get_json()
    else:
        data = {}

    try:
        download_folder, file_name = (
            self.manager.export_capabilities_and_policies(id=id, **data))
    except exception.ViggoCoreException as exc:
        return flask.Response(response=exc.message,
                              status=exc.status)

    response = flask.send_from_directory(download_folder, file_name)
    utils.remove_file(f'{download_folder}/{file_name}')
    return response



from sqlalchemy import func, or_, and_
from viggocore.subsystem.application.resource import Application
from typing import List
from viggocore.common import exception
from viggocore.common.input import RouteResource, InputResource, \
    InputResourceUtils
from viggocore.common.subsystem import operation, manager
from viggocore.subsystem.capability.resource import Capability
from viggocore.subsystem.capability_functionality.resource import (
    CapabilityFunctionality)
from viggocore.subsystem.capability_module.resource import CapabilityModule
from viggocore.subsystem.policy.resource import Policy
from viggocore.subsystem.policy_functionality.resource import (
    PolicyFunctionality)
from viggocore.subsystem.policy_module.resource import PolicyModule
from viggocore.subsystem.role.resource import Role
from viggocore.subsystem.route.resource import Route
from viggocore.common.subsystem.pagination import Pagination
from sqlalchemy.sql import text
from viggocore.subsystem.application.functions.manager import (
    export_capabilities_and_policies,
    export_capabilities_and_policies_sql,
    get_xlsx_import_model,
    import_capabilities_and_policies)


class Create(operation.Create):

    def pre(self, session, **kwargs) -> bool:
        self.exceptions = kwargs.pop('exceptions', [])

        return super().pre(session, **kwargs)

    def do(self, session, **kwargs):
        super().do(session)
        self.manager.create_user_capabilities_and_policies(id=self.entity.id,
                                                           session=session)
        if self.entity.name != Application.DEFAULT:
            self.manager.create_admin_capabilities_and_policies(
                id=self.entity.id, session=session, exceptions=self.exceptions)
        return self.entity


class ListManager(operation.List):

    def apply_filters(self, query, resource, **kwargs):
        for k, v in kwargs.items():
            if hasattr(resource, k):
                if k == 'tag':
                    values = v
                    if len(v) > 0 and v[0] == '#':
                        values = v[1:]
                    values = values.split(',')
                    filter_tags = []
                    for value in values:
                        filter_tags.append(
                            getattr(resource, k)
                            .like('%#' + str(value) + ' %'))
                    query = query.filter(or_(*filter_tags))
                elif k == 'tag_name':
                    values = v
                    if len(v) > 0 and v[0] == '#':
                        values = v[1:]
                    filter_tags = []
                    query = query.filter(
                        getattr(resource, k) == '#' + str(values) + ' ')
                elif isinstance(v, str) and '%' in v:
                    normalize = func.viggocore_normalize
                    query = query.filter(normalize(getattr(resource, k))
                                         .ilike(normalize(v)))
                else:
                    query = query.filter(getattr(resource, k) == v)

        return query

    def apply_pagination(self, query, pagination: Pagination):
        if (pagination.order_by is not None and pagination.page is not None
           and pagination.page_size is not None):
            query = query.order_by(text(pagination.order_by))

        if pagination.page_size is not None:
            query = query.limit(pagination.page_size)
            if pagination.page is not None:
                query = query.offset(pagination.page * pagination.page_size)

        return query

    def do(self, session, **kwargs):
        not_default = kwargs.pop('not_default', False)
        query = session.query(Application)

        pagination = Pagination.get_pagination(Application, **kwargs)

        if not_default is True:
            query = query.filter(Application.name != Application.DEFAULT)

        query = self.apply_filters(query, Application, **kwargs)
        query = self.apply_pagination(query, pagination)

        result = query.all()
        return result


class CreateUserCapabilitiesAndPolicies(operation.Operation):

    def pre(self, session, id, **kwargs) -> bool:
        self.application_id = id
        self.user_resources = self.manager.bootstrap_resources.USER
        self.role_id = self.manager.api.roles().\
            get_role_by_name(role_name=Role.USER).id

        return True

    def do(self, session, **kwargs):
        self.resources = {'resources': self.user_resources}
        self.manager.api.capabilities().create_capabilities(
            id=self.application_id, **self.resources)

        self.resources['application_id'] = self.application_id
        self.manager.api.roles().create_policies(id=self.role_id,
                                                 **self.resources)


class CreateAdminCapabilitiesAndPolicies(operation.Operation):

    def _map_routes(self, routes: List[Route]) -> List[RouteResource]:
        resources = [(route.url, route.method) for route in routes]
        return resources

    def _filter_resources(self, all_resources: List[RouteResource],
                          exceptions_resources: List[RouteResource],
                          sysadmin_exclusive_resources: List[RouteResource],
                          user_resources: List[RouteResource]) \
            -> List[RouteResource]:
        resources = InputResourceUtils.diff_resources(
            all_resources, sysadmin_exclusive_resources, user_resources,
            exceptions_resources)
        return resources

    def pre(self, session, id, exceptions: List[InputResource] = [], **kwargs):
        self.application_id = id
        exceptions_resources = InputResourceUtils.parse_resources(exceptions)

        routes = self.manager.api.routes().list(active=True)
        routes_resources = self._map_routes(routes)

        self.admin_role_id = self.manager.api.roles().\
            get_role_by_name(role_name=Role.ADMIN).id

        self.admin_resources = self._filter_resources(
            routes_resources, exceptions_resources,
            self.manager.bootstrap_resources.SYSADMIN_EXCLUSIVE,
            self.manager.bootstrap_resources.USER)
        return True

    def do(self, session, **kwargs):
        self.resources = {'resources': self.admin_resources}
        self.manager.api.capabilities().create_capabilities(
            id=self.application_id, **self.resources)

        self.resources['application_id'] = self.application_id
        self.manager.api.roles().create_policies(id=self.admin_role_id,
                                                 **self.resources)


class CreateCapabilitiesWithExceptions(operation.Operation):

    def _filter_resources(self, routes: List[Route],
                          exceptions: List[RouteResource],
                          sysadmin_exclusive_resources: List[RouteResource],
                          user_resources: List[RouteResource]) \
            -> List[RouteResource]:
        all_resources = [(route.url, route.method) for route in routes]
        resources = InputResourceUtils.diff_resources(
            all_resources, sysadmin_exclusive_resources, user_resources,
            exceptions)
        return resources

    def pre(self, session, id: str, **kwargs):
        self.application_id = id
        exceptions = kwargs.get('exceptions', None)

        if not self.application_id or exceptions is None:
            raise exception.BadRequest()

        routes = self.manager.api.routes().list(active=True)
        exceptions_resources = InputResourceUtils.parse_resources(exceptions)
        self.resources = self.\
            _filter_resources(routes,
                              exceptions_resources,
                              self.manager.bootstrap_resoures.
                              SYSADMIN_EXCLUSIVE,
                              self.manager.bootstrap_resources.USER)

        return self.driver.get(id, session=session) is not None

    def do(self, session, **kwargs):
        data = {'resources', self.resources}
        self.manager.api.capabilities().\
            create_capabilities(id=self.application_id, **data)


class GetRoles(operation.Operation):

    def pre(self, session, id, **kwargs):
        self.application_id = id
        return self.driver.get(id, session=session) is not None

    def do(self, session, **kwargs):
        roles = session.query(Role). \
            join(Policy). \
            join(Capability). \
            filter(and_(Capability.application_id == self.application_id,
                        Role.name != Role.USER)). \
            distinct().all()

        roles_module = session.query(Role). \
            join(PolicyModule, Role.id == PolicyModule.role_id). \
            join(CapabilityModule,
                 CapabilityModule.id == PolicyModule.capability_module_id). \
            filter(and_(CapabilityModule.application_id == self.application_id,
                        Role.name != Role.USER)). \
            distinct().all()

        roles_functionality = session.query(Role). \
            join(PolicyFunctionality,
                 Role.id == PolicyFunctionality.role_id). \
            join(CapabilityFunctionality,
                 CapabilityFunctionality.id ==
                 PolicyFunctionality.capability_functionality_id). \
            filter(and_(CapabilityFunctionality.application_id ==
                        self.application_id,
                        Role.name != Role.USER)). \
            distinct().all()

        roles += roles_module
        roles += roles_functionality
        return set(roles)


class UpdateSettings(operation.Update):

    def pre(self, session, id: str, **kwargs) -> bool:
        self.settings = kwargs
        if self.settings is None or not self.settings:
            raise exception.BadRequest("Erro! There is not a setting")
        return super().pre(session=session, id=id)

    def do(self, session, **kwargs):
        result = {}
        for key, value in self.settings.items():
            new_value = self.entity.update_setting(key, value)
            result[key] = new_value
        super().do(session)

        return result


class RemoveSettings(operation.Update):

    def pre(self, session, id: str, **kwargs) -> bool:
        self.keys = kwargs.get('keys', [])
        if not self.keys:
            raise exception.BadRequest('Erro! keys are empty')
        super().pre(session, id=id)

        return self.entity.is_stable()

    def do(self, session, **kwargs):
        result = {}
        for key in self.keys:
            value = self.entity.remove_setting(key)
            result[key] = value
        super().do(session=session)

        return result


class GetApplicationSettingsByKeys(operation.Get):

    def pre(self, session, id, **kwargs):
        self.keys = kwargs.get('keys', [])
        if not self.keys:
            raise exception.BadRequest('Erro! keys are empty')
        return super().pre(session, id=id)

    def do(self, session, **kwargs):
        entity = super().do(session=session)
        settings = {}
        for key in self.keys:
            value = entity.settings.get(key, None)
            if value is not None:
                settings[key] = value
        return settings


class Manager(manager.Manager):

    def __init__(self, driver):
        super().__init__(driver)
        self.create = Create(self)
        self.list = ListManager(self)
        self.create_user_capabilities_and_policies = \
            CreateUserCapabilitiesAndPolicies(self)
        self.create_admin_capabilities_and_policies = \
            CreateAdminCapabilitiesAndPolicies(self)
        self.get_roles = GetRoles(self)
        self.update_settings = UpdateSettings(self)
        self.remove_settings = RemoveSettings(self)
        self.get_application_settings_by_keys = \
            GetApplicationSettingsByKeys(self)
        self.export_capabilities_and_policies_sql = (
            export_capabilities_and_policies_sql
            .ExportCapabilitiesAndPoliciesSql(self))
        self.export_capabilities_and_policies = (
            export_capabilities_and_policies
            .ExportCapabilitiesAndPolicies(self))
        self.import_capabilities_and_policies = (
            import_capabilities_and_policies
            .ImportCapabilitiesAndPolicies(self))
        self.get_xlsx_import_model = (
            get_xlsx_import_model.GetXlsxImportModel(self))



from viggocore.common import exception, utils as core_utils
from viggocore.subsystem.capability.resource import Capability
from viggocore.subsystem.role.resource import Role
from viggocore.subsystem.route.resource import Route
from viggocore.subsystem.policy.resource import Policy
from sqlalchemy import func, or_
from openpyxl.worksheet.worksheet import Worksheet
from openpyxl import Workbook, styles
from viggocore.common.subsystem.manager_functions import make_exportar_xlsx
from datetime import datetime as datetime1


class ExportCapabilitiesAndPolicies(make_exportar_xlsx.MakeExportarXsls):

    # funções sobrescritas
    def _formatar_celulas(self, ws, dados, cols, is_to_bold=True):
        qtd_linhas = len(dados)

        if qtd_linhas > 0:
            lin_ini = 2
            lin_fim = lin_ini + qtd_linhas  # A última linha é de total

            for lin in range(lin_ini, lin_fim + 1):
                # Accounting format
                for col in cols:
                    cell = ws.cell(lin, col['num'])
                    cell.border = styles.Border(
                        top=styles.Side(border_style='thin'),
                        right=styles.Side(border_style='thin'),
                        bottom=styles.Side(border_style='thin'),
                        left=styles.Side(border_style='thin'))

                    if 'fmt' in col:
                        cell.number_format = col['fmt']

                    if 'align' in col:
                        cell.alignment = styles.Alignment(
                            horizontal=col['align'])

                    if 'function' in col:
                        formula = col['function'] \
                            .replace('[row]', str(cell.row))
                        cell.value = formula

                    if (lin == lin_fim):
                        self._formatar_footer(
                            col, cell, lin_ini, qtd_linhas)
        return

    # função para pegar o nome do arquivo
    def _get_file_name(self, **kwargs):
        file_name = kwargs.get('file_name', None)
        if file_name is not None:
            return file_name
        else:
            application = self.manager.get(id=self.id)

            file_name = f'{application.name}-{self.a_partir}.xlsx'
            return file_name

    # valida se o tipo de exportacao passado é válido
    def _validar_tipo_exportacao(self):
        tipos_permitidos = ['CAPACIDADE', 'POLITICA']
        if any(list(map(
                lambda x: x not in tipos_permitidos, self.tipo_exportacao))):
            raise exception.BadRequest(
                'Os tipos de exportação permitidos são:' +
                ', '.join(tipos_permitidos) + '.')

    # funções de buscar dados
    def _get_capabilities(self, session):
        query = session.query(Route.url, Route.method)\
            .join(Capability, Capability.route_id == Route.id)\
            .filter(Capability.application_id == self.id)\
            .filter(or_(Route.created_at >= self.a_partir,
                        Route.created_at.is_(None)))\
            .order_by(Route.url, Route.method)
        result = query.all()
        return result

    def _get_policies(self, session):
        papeis = [p.upper() for p in self.papeis]
        query = session.query(Role.name, Route.url, Route.method)\
            .join(Policy, Policy.role_id == Role.id)\
            .join(Capability, Capability.id == Policy.capability_id)\
            .join(Route, Route.id == Capability.route_id)\
            .filter(func.upper(Role.name).in_(papeis))\
            .filter(Capability.application_id == self.id)\
            .filter(or_(Route.created_at >= self.a_partir,
                        Route.created_at.is_(None)))\
            .order_by(Role.name, Route.url, Route.method)
        result = query.all()
        return result

    def _get_capabilities_and_policies(self, session):
        # inicia as listas de capacidades e políticas
        capabilities = []
        policies = []

        # se foi passada a CAPACIDADE como tipo_exportacao então vai buscar as
        # capacidades
        if 'CAPACIDADE' in self.tipo_exportacao:
            capabilities = self._get_capabilities(session=session)

        # se foi passada a POLITICA como tipo_exportacao então vai buscar as
        # políticas ligadas as capacidades desta aplicação
        if 'POLITICA' in self.tipo_exportacao:
            policies = self._get_policies(session=session)

        return (capabilities, policies)

    # funções de preenchimento do xlsx
    def _gen_work_sheet_capabilities(self, ws: Worksheet, capabilities):
        ROW_INIT = 1

        cols = [
            {'num': 1, 'name': 'URL', 'width': 50},
            {'num': 2, 'name': 'METHOD', 'width': 25}
        ]

        self._def_header_style(ws, cols, 'export_capabilities',
                               row_ini=ROW_INIT)

        for linha in capabilities:
            ws.append(list(linha))

        self._formatar_celulas(ws, capabilities, cols)
        self._style_header(ws, cols)
        self._align_horizontal_center_column(ws, capabilities, 2)

        return True

    def _gen_work_sheet_policies(self, ws: Worksheet, policies):
        ROW_INIT = 1

        cols = [
            {'num': 1, 'name': 'PAPEL', 'width': 25},
            {'num': 2, 'name': 'URL', 'width': 50},
            {'num': 3, 'name': 'METHOD', 'width': 25}
        ]

        self._def_header_style(ws, cols, 'export_policies',
                               row_ini=ROW_INIT)

        for linha in policies:
            ws.append(list(linha))

        self._formatar_celulas(ws, policies, cols)
        self._style_header(ws, cols)
        self._align_horizontal_center_column(ws, policies, 2)

        return True

    def pre(self, id, **kwargs):
        self.id = id
        if self.id is None:
            raise exception.BadRequest(
                'É obrigatório passar o id na rota da requisição.')

        self.tipo_exportacao = kwargs.pop('tipo_exportacao', [])
        if not self.tipo_exportacao:
            raise exception.BadRequest(
                'É obrigatório informar o tipo_exportacao.')

        self._validar_tipo_exportacao()

        self.papeis = kwargs.pop('papeis', [])
        if 'POLITICA' in self.tipo_exportacao and not self.papeis:
            raise exception.BadRequest(
                'É obrigatório informar os papéis caso você ' +
                'queira exportar as políticas também.')

        self.a_partir = kwargs.pop('a_partir', None)
        if self.a_partir is None:
            raise exception.BadRequest(
                'É obrigatório informar o campo "a_partir" para filtrar ' +
                'as informações que foram geradas posterior a esta data.')
        else:
            try:
                self.a_partir = datetime1.strptime(
                    self.a_partir.replace(' ', '+'), '%Y-%m-%d%z')
            except Exception:
                raise exception.BadRequest(
                    'É obrigatório passar a data no formato "yyyy-mm-dd-tz."')

        return True

    def do(self, session, **kwargs):
        capabilities, policies = self._get_capabilities_and_policies(
            session=session)

        if len(capabilities) == 0 and len(policies) == 0:
            raise exception.BadRequest(
                'Não existe dados para essa filtragem.')

        self.file_name = self._get_file_name(**kwargs)
        self.download_folder = core_utils.get_upload_folder()

        wb = Workbook()

        # Aba de capacidades
        ws_cap = wb.active  # Usa a aba ativa pra primeira
        ws_cap.title = "CAPACIDADES"
        self._gen_work_sheet_capabilities(ws_cap, capabilities)

        # Aba de políticas
        ws_pol = wb.create_sheet(title="POLITICAS")
        self._gen_work_sheet_policies(ws_pol, policies)

        wb.save(filename=self.download_folder + '/' + self.file_name)

        return self.download_folder, self.file_name



from viggocore.common.subsystem import router


class Router(router.Router):

    def __init__(self, collection, routes=[]):
        super().__init__(collection, routes)

    @property
    def routes(self):
        settings_endpoint = '/settings'
        return super().routes + [
            {
                'action': 'getApplicationRoles',
                'method': 'GET',
                'url': self.resource_url + '/roles',
                'callback': 'get_roles'
            },
            {
                'action': 'Update settings on Application',
                'method': 'PUT',
                'url': self.resource_url + settings_endpoint,
                'callback': 'update_settings',
                'bypass': False
            },
            {
                'action': 'Remove settings from Application',
                'method': 'DELETE',
                'url': self.resource_url + settings_endpoint,
                'callback': 'remove_settings',
                'bypass': False
            },
            {
                'action': 'Get settings by keys from Application',
                'method': 'GET',
                'url': self.resource_url + settings_endpoint,
                'callback': 'get_application_settings_by_keys',
                'bypass': False
            },
            {
                'action': 'Exports application capabilities and policies',
                'method': 'POST',
                'url': self.resource_url + '/export_capabilities_and_policies',
                'callback': 'export_capabilities_and_policies',
                'bypass': False
            },
            {
                'action': 'Import application capabilities and policies',
                'method': 'POST',
                'url': self.resource_url + '/import_capabilities_and_policies',
                'callback': 'import_capabilities_and_policies',
                'bypass': False
            },
            {
                'action': ('Exports application capabilities and ' +
                           'policies in SQL'),
                'method': 'POST',
                'url': (self.resource_url +
                        '/export_capabilities_and_policies_sql'),
                'callback': 'export_capabilities_and_policies_sql',
                'bypass': False
            },
            {
                'action': 'Exports application capabilities and policies',
                'method': 'GET',
                'url': self.collection_url + '/xlsx_import_model',
                'callback': 'get_xlsx_import_model',
                'bypass': False
            },
        ]
